""" Tests for ast_utils """
import ast
import pickle
from ast import (
    AnnAssign,
    Assign,
    Attribute,
    BinOp,
    Call,
    ClassDef,
    Constant,
    Dict,
    Expr,
    FunctionDef,
    Import,
    ImportFrom,
    List,
    Load,
    Module,
    Mult,
    Name,
    NameConstant,
    Store,
    Str,
    Tuple,
    arg,
    arguments,
    keyword,
)
from copy import deepcopy
from os import path
from unittest import TestCase

from cdd.ast_utils import (
    NoneStr,
    RewriteAtQuery,
    _parse_default_from_ast,
    annotate_ancestry,
    cmp_ast,
    emit_ann_assign,
    emit_arg,
    find_ast_type,
    find_in_ast,
    get_at_root,
    get_function_type,
    get_value,
    infer_type_and_default,
    maybe_type_comment,
    node_to_dict,
    param2argparse_param,
    param2ast,
    parse_to_scalar,
    set_arg,
    set_docstring,
    set_slice,
    set_value,
)
from cdd.pure_utils import PY3_8, PY_GTE_3_8, tab
from cdd.source_transformer import ast_parse
from cdd.tests.mocks.argparse import argparse_add_argument_expr
from cdd.tests.mocks.classes import class_ast, class_str
from cdd.tests.mocks.doctrans import function_type_annotated
from cdd.tests.mocks.methods import (
    class_with_method_and_body_types_ast,
    class_with_method_and_body_types_str,
    class_with_optional_arg_method_ast,
    function_adder_str,
)
from cdd.tests.utils_for_tests import inspectable_compile, run_ast_test, unittest_main


class TestAstUtils(TestCase):
    """Test class for ast_utils"""

    def test_annotate_ancestry(self) -> None:
        """Tests that `annotate_ancestry` properly decorates"""
        node = Module(
            body=[
                AnnAssign(
                    annotation=Name(
                        "str",
                        Load(),
                    ),
                    simple=1,
                    target=Name("dataset_name", Store()),
                    value=set_value("~/tensorflow_datasets"),
                    expr=None,
                    expr_target=None,
                    expr_annotation=None,
                ),
                Assign(
                    annotation=None,
                    simple=1,
                    targets=[Name("epochs", Store())],
                    value=set_value("333"),
                    expr=None,
                    expr_target=None,
                    expr_annotation=None,
                    **maybe_type_comment
                ),
            ],
            stmt=None,
        )
        self.assertFalse(hasattr(node.body[0], "_location"))
        self.assertFalse(hasattr(node.body[1], "_location"))
        annotate_ancestry(node)
        self.assertEqual(node.body[0]._location, ["dataset_name"])
        self.assertEqual(node.body[1]._location, ["epochs"])

    def test_cmp_ast(self) -> None:
        """Test `cmp_ast` branch that isn't tested anywhere else"""
        self.assertFalse(cmp_ast(None, 5))

    def test_emit_ann_assign(self) -> None:
        """Tests that AnnAssign is emitted from `emit_ann_assign`"""
        self.assertIsInstance(class_ast.body[1], AnnAssign)
        self.assertIsInstance(emit_ann_assign(class_ast.body[1]), AnnAssign)
        self.assertIsInstance(emit_ann_assign(class_ast.body[1]), AnnAssign)
        gen_ast = emit_ann_assign(
            find_in_ast(
                "C.function_name.dataset_name".split("."),
                class_with_method_and_body_types_ast,
            )
        )
        self.assertIsInstance(gen_ast, AnnAssign)
        run_ast_test(
            self,
            gen_ast,
            AnnAssign(
                annotation=Name(
                    "str",
                    Load(),
                ),
                simple=1,
                target=Name("dataset_name", Store()),
                value=set_value("~/tensorflow_datasets"),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )

    def test_emit_ann_assign_fails(self) -> None:
        """Tests that `emit_ann_assign` raised the right error"""
        for typ_ in FunctionDef, ClassDef, Module:
            self.assertRaises(NotImplementedError, lambda: emit_ann_assign(typ_))

    def test_emit_arg(self) -> None:
        """Tests that `arg` is emitted from `emit_arg`"""
        self.assertIsInstance(
            class_with_method_and_body_types_ast.body[1].args.args[1], arg
        )
        self.assertIsInstance(
            emit_arg(class_with_method_and_body_types_ast.body[1].args.args[1]), arg
        )
        assign = Assign(
            targets=[Name("yup", Store())],
            value=set_value("nup"),
            expr=None,
            **maybe_type_comment
        )
        gen_ast = emit_arg(assign)
        self.assertIsInstance(gen_ast, arg)
        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=set_arg("yup"),
        )

    def test_emit_arg_fails(self) -> None:
        """Tests that `emit_arg` raised the right error"""
        for typ_ in FunctionDef, ClassDef, Module:
            self.assertRaises(NotImplementedError, lambda: emit_arg(typ_))

    def test_find_in_ast(self) -> None:
        """Tests that `find_in_ast` successfully finds nodes in AST"""

        run_ast_test(
            self,
            find_in_ast("ConfigClass.dataset_name".split("."), class_ast),
            AnnAssign(
                annotation=Name(
                    "str",
                    Load(),
                ),
                simple=1,
                target=Name("dataset_name", Store()),
                value=set_value("mnist"),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )

    def test_find_in_ast_self(self) -> None:
        """Tests that `find_in_ast` successfully finds itself in AST"""
        run_ast_test(self, find_in_ast(["ConfigClass"], class_ast), class_ast)
        module = Module(body=[], type_ignores=[], stmt=None)
        run_ast_test(self, find_in_ast([], module), module)
        module_with_fun = Module(
            body=[
                FunctionDef(
                    name="call_peril",
                    args=arguments(
                        args=[],
                        defaults=[],
                        kw_defaults=[],
                        kwarg=None,
                        kwonlyargs=[],
                        posonlyargs=[],
                        vararg=None,
                        arg=None,
                    ),
                    body=[],
                    decorator_list=[],
                    lineno=None,
                    arguments_args=None,
                    identifier_name=None,
                    stmt=None,
                )
            ],
            stmt=None,
        )
        annotate_ancestry(module_with_fun)
        run_ast_test(
            self,
            find_in_ast(["call_peril"], module_with_fun),
            module_with_fun.body[0],
            skip_black=True,
        )

    def test_find_in_ast_None(self) -> None:
        """Tests that `find_in_ast` fails correctly in AST"""
        self.assertIsNone(find_in_ast(["John Galt"], class_ast))

    def test_find_in_ast_no_val(self) -> None:
        """Tests that `find_in_ast` correctly gives AST node from
        `def class C(object): def function_name(self,dataset_name: str,…)`"""
        run_ast_test(
            self,
            find_in_ast(
                "C.function_name.dataset_name".split("."),
                class_with_optional_arg_method_ast,
            ),
            set_arg(
                annotation=Name(
                    "str",
                    Load(),
                ),
                arg="dataset_name",
            ),
        )

    def test_find_in_ast_with_val(self) -> None:
        """Tests that `find_in_ast` correctly gives AST node from
        `def class C(object): def function_name(self,dataset_name: str='foo',…)`"""
        gen_ast = find_in_ast(
            "C.function_name.dataset_name".split("."),
            class_with_method_and_body_types_ast,
        )

        self.assertIsInstance(gen_ast.default, Constant if PY_GTE_3_8 else Str)

        self.assertEqual(get_value(gen_ast.default), "~/tensorflow_datasets")
        run_ast_test(
            self,
            gen_ast,
            set_arg(
                annotation=Name(
                    "str",
                    Load(),
                ),
                arg="dataset_name",
            ),
        )

    def test_get_at_root(self) -> None:
        """Tests that `get_at_root` successfully gets the imports"""
        with open(path.join(path.dirname(__file__), "mocks", "eval.py")) as f:
            imports = get_at_root(ast.parse(f.read()), (Import, ImportFrom))
        self.assertIsInstance(imports, list)
        self.assertEqual(len(imports), 1)

        self.assertTrue(
            cmp_ast(
                imports[0],
                ast.Import(
                    names=[
                        ast.alias(
                            asname=None,
                            name="cdd.tests.mocks",
                            identifier=None,
                            identifier_name=None,
                        )
                    ],
                    alias=None,
                ),
            )
        )

    def test_node_to_dict(self) -> None:
        """
        Tests `node_to_dict`
        """
        self.assertDictEqual(
            node_to_dict(set_arg(arg="a", annotation=Name("int", Load()))),
            dict(
                annotation="int",
                arg="a",
                **{"expr": None, "type_comment": None} if PY_GTE_3_8 else {}
            ),
        )

    def test_replace_in_ast_with_val(self) -> None:
        """
        Tests that `RewriteAtQuery` can actually replace a node at given location
        """
        parsed_ast = ast_parse(class_with_method_and_body_types_str)
        rewrite_at_query = RewriteAtQuery(
            search="C.function_name.dataset_name".split("."),
            replacement_node=AnnAssign(
                annotation=Name("int", Load()),
                simple=1,
                target=Name("dataset_name", Store()),
                value=set_value(15),
                expr=None,
                expr_annotation=None,
                expr_target=None,
            ),
        )
        gen_ast = rewrite_at_query.visit(parsed_ast)
        self.assertTrue(rewrite_at_query.replaced, True)

        run_ast_test(
            self,
            gen_ast,
            ast.parse(
                class_with_method_and_body_types_str.replace(
                    'dataset_name: str = "mnist"', "dataset_name: int = 15"
                )
            ),
        )

    def test_replace_in_ast_with_val_on_non_function(self) -> None:
        """
        Tests that `RewriteAtQuery` can actually replace a node at given location
        """
        parsed_ast = ast_parse(class_str)
        rewrite_at_query = RewriteAtQuery(
            search="ConfigClass.dataset_name".split("."),
            replacement_node=AnnAssign(
                annotation=Name("int", Load()),
                simple=1,
                target=Name("dataset_name", Store()),
                value=set_value(15),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )
        gen_ast = rewrite_at_query.visit(parsed_ast)
        self.assertTrue(rewrite_at_query.replaced, True)

        run_ast_test(
            self,
            gen_ast,
            ast.parse(
                class_str.replace(
                    'dataset_name: str = "mnist"', "dataset_name: int = 15"
                )
            ),
        )

    def test_get_function_type(self) -> None:
        """Test get_function_type returns the right type"""
        self.assertEqual(
            get_function_type(
                FunctionDef(
                    args=arguments(
                        args=[set_arg("something else")],
                        arg=None,
                    ),
                    arguments_args=None,
                    identifier_name=None,
                    stmt=None,
                )
            ),
            "static",
        )
        self.assertEqual(
            get_function_type(
                FunctionDef(
                    args=arguments(args=[], arg=None),
                    arguments_args=None,
                    identifier_name=None,
                    stmt=None,
                )
            ),
            "static",
        )
        self.assertEqual(
            get_function_type(
                FunctionDef(
                    args=arguments(
                        args=[set_arg("self")],
                        arg=None,
                    ),
                    arguments_args=None,
                    identifier_name=None,
                    stmt=None,
                )
            ),
            "self",
        )
        self.assertEqual(
            get_function_type(
                FunctionDef(
                    args=arguments(
                        args=[set_arg("cls")],
                        arg=None,
                    ),
                    arguments_args=None,
                    identifier_name=None,
                    stmt=None,
                )
            ),
            "cls",
        )

    def test_get_value(self) -> None:
        """Tests get_value succeeds"""
        val = "foo"
        self.assertEqual(get_value(Str(s=val, constant_value=None, string=None)), val)
        self.assertEqual(
            get_value(Constant(value=val, constant_value=None, string=None)), val
        )
        self.assertIsInstance(get_value(Tuple(expr=None)), Tuple)
        self.assertIsInstance(get_value(Tuple(expr=None)), Tuple)
        self.assertIsNone(get_value(Name(None, None)))
        self.assertEqual(get_value(get_value(ast.parse("-5").body[0])), -5)

    def test_set_value(self) -> None:
        """Tests that `set_value` returns the right type for the right Python version"""
        import cdd.ast_utils

        _cdd_ast_utils_PY3_8_orig = PY3_8
        # patch stopped working, getattr failed also ^
        try:
            cdd.ast_utils.PY3_8 = True

            self.assertIsInstance(
                cdd.ast_utils.set_value(None),
                Constant if PY_GTE_3_8 else NameConstant,
            )

            cdd.ast_utils.PY3_8 = False

            self.assertIsInstance(cdd.ast_utils.set_value(None), NameConstant)
            self.assertIsInstance(cdd.ast_utils.set_value(True), NameConstant)

            cdd.ast_utils.PY3_8 = True
            self.assertIsInstance(
                cdd.ast_utils.set_value("foo"),
                Constant if PY_GTE_3_8 else Str,
            )

            cdd.ast_utils.PY3_8 = False
            self.assertIsInstance(cdd.ast_utils.set_value("foo"), Str)
        finally:
            cdd.ast_utils = _cdd_ast_utils_PY3_8_orig

    def test_set_docstring(self) -> None:
        """
        Tests that `set_docstring` sets the docstring
        """
        with_doc_str = deepcopy(function_type_annotated)
        doc_str = ast.get_docstring(ast.parse(function_adder_str))
        set_docstring(doc_str, False, with_doc_str)
        self.assertIsNone(ast.get_docstring(function_type_annotated))
        self.assertEqual(ast.get_docstring(with_doc_str), doc_str)

        without_doc_str = deepcopy(function_type_annotated)
        doc_str = "\t\n"
        set_docstring(doc_str, False, without_doc_str)
        self.assertIsNone(ast.get_docstring(without_doc_str), doc_str)

    def test_find_ast_type(self) -> None:
        """Test that `find_ast_type` gives the wrapped class back"""

        class_def = ClassDef(
            name="",
            bases=tuple(),
            keywords=tuple(),
            decorator_list=[],
            body=[],
            expr=None,
            identifier_name=None,
        )
        run_ast_test(
            self,
            find_ast_type(Module(body=[class_def], stmt=None)),
            class_def,
            skip_black=True,
        )

    def test_find_ast_type_fails(self) -> None:
        """Test that `find_ast_type` throws the right errors"""

        self.assertRaises(NotImplementedError, lambda: find_ast_type(None))
        self.assertRaises(NotImplementedError, lambda: find_ast_type(""))
        self.assertRaises(TypeError, lambda: find_ast_type(Module(body=[], stmt=None)))
        self.assertRaises(
            NotImplementedError,
            lambda: find_ast_type(
                Module(
                    body=[
                        ClassDef(expr=None, identifier_name=None),
                        ClassDef(expr=None, identifier_name=None),
                    ],
                    stmt=None,
                )
            ),
        )

    def test_to_named_class_def(self) -> None:
        """Test that find_ast_type gives the wrapped named class back"""

        class_def = ClassDef(
            name="foo",
            bases=tuple(),
            keywords=tuple(),
            decorator_list=[],
            body=[],
            expr=None,
            identifier_name=None,
        )
        run_ast_test(
            self,
            find_ast_type(
                Module(
                    body=[
                        ClassDef(
                            name="bar",
                            bases=tuple(),
                            keywords=tuple(),
                            decorator_list=[],
                            body=[],
                            expr=None,
                            identifier_name=None,
                        ),
                        class_def,
                    ],
                    stmt=None,
                ),
                node_name="foo",
            ),
            class_def,
            skip_black=True,
        )

    def test_param2ast_with_assign(self) -> None:
        """Check that `param2ast` behaves correctly with a non annotated (typeless) input"""

        run_ast_test(
            self,
            param2ast(
                ("zion", {"typ": None}),
            ),
            gold=AnnAssign(
                annotation=Name("object", Load()),
                simple=1,
                target=Name("zion", Store()),
                value=set_value(None),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )

    def test_param2ast_with_assign_dict(self) -> None:
        """Check that `param2ast` behaves correctly with dict type"""

        run_ast_test(
            self,
            param2ast(
                ("menthol", {"typ": "dict"}),
            ),
            gold=AnnAssign(
                annotation=set_slice(Name("dict", Load())),
                simple=1,
                target=Name("menthol", Store()),
                value=Dict(keys=[], values=[], expr=None),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )

    def test_param2ast_with_bad_default(self) -> None:
        """Check that `param2ast` behaves correctly with a bad default"""

        run_ast_test(
            self,
            param2ast(
                (
                    "stateful_metrics",
                    {"typ": "NoneType", "default": "the `Model`'s metrics"},
                ),
            ),
            gold=AnnAssign(
                annotation=Name("NoneType", Load()),
                simple=1,
                target=Name("stateful_metrics", Store()),
                value=set_value("```the `Model`'s metrics```"),
                expr=None,
                expr_annotation=None,
                expr_target=None,
            ),
        )

    def test_param2ast_with_wrapped_default(self) -> None:
        """Check that `param2ast` behaves correctly with a wrapped default"""

        run_ast_test(
            self,
            param2ast(
                ("zion", {"typ": None, "default": set_value(NoneStr)}),
            ),
            gold=AnnAssign(
                annotation=Name("object", Load()),
                simple=1,
                target=Name("zion", Store()),
                value=set_value(None),
                expr=None,
                expr_target=None,
                expr_annotation=None,
            ),
        )

    def test_param2argparse_param_none_default(self) -> None:
        """
        Tests that param2argparse_param works to reparse the default
        """
        run_ast_test(
            gen_ast=param2argparse_param(("yup", {"default": NoneStr})),
            gold=Expr(
                Call(
                    args=[set_value("--yup")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_simple_type(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                ("byo", {"default": 5, "typ": "str"}),
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(arg="type", value=Name("int", Load()), identifier=None),
                        keyword(arg="required", value=set_value(True), identifier=None),
                        keyword(arg="default", value=set_value(5), identifier=None),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_list(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a list
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                ("byo", {"default": [], "typ": "str"}),
            ),
            gold=argparse_add_argument_expr,
            test_case_instance=self,
        )

    def test_param2argparse_param_default_ast_tuple(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is an ast.Tuple
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": Tuple(
                            elts=[],
                            ctx=Load(),
                            expr=None,
                        ),
                        "typ": "str",
                    },
                ),
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(
                            arg="type", value=Name("loads", Load()), identifier=None
                        ),
                        keyword(arg="required", value=set_value(True), identifier=None),
                        keyword(arg="default", value=set_value("()"), identifier=None),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_ast_list(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is an ast.List
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": List(
                            elts=[],
                            ctx=Load(),
                            expr=None,
                        ),
                        "typ": "str",
                    },
                ),
            ),
            gold=argparse_add_argument_expr,
            test_case_instance=self,
        )

    def test_param2argparse_param_default_ast_expr_with_list(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is an ast.List inside an ast.Expr
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": Expr(
                            List(
                                elts=[],
                                ctx=Load(),
                                expr=None,
                            ),
                            expr_value=None,
                        ),
                        "typ": "str",
                    },
                ),
            ),
            gold=argparse_add_argument_expr,
            test_case_instance=self,
        )

    def test_param2argparse_param_default_ast_binop(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a non specially handled ast.AST
        """
        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": BinOp(
                            set_value(5),
                            Mult(),
                            set_value(5),
                        ),
                        "typ": "str",
                    },
                ),
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(arg="required", value=set_value(True), identifier=None),
                        keyword(
                            arg="default",
                            value=set_value("```(5 * 5)```"),
                            identifier=None,
                        ),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_function(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is an in-memory function
        """

        function_str = (
            "from operator import add\n"
            "def adder(a, b):\n"
            "{tab}return add(a, b)".format(tab=tab)
        )
        adder = getattr(
            inspectable_compile(function_str),
            "adder",
        )
        pickled_adder = pickle.dumps(adder)  # eww

        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": adder,
                        "typ": "str",
                    },
                ),
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(
                            arg="type",
                            value=Name("pickle.loads", Load()),
                            identifier=None,
                        ),
                        keyword(
                            arg="default",
                            value=set_value(pickled_adder),
                            identifier=None,
                        ),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_code_quoted(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a code quoted str
        """
        run_ast_test(
            gen_ast=(
                param2argparse_param(
                    (
                        "byo",
                        {
                            "default": "```(4)```",
                            "typ": "str",
                        },
                    ),
                )
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(arg="type", value=Name("int", Load()), identifier=None),
                        keyword(arg="required", value=set_value(True), identifier=None),
                        keyword(arg="default", value=set_value(4), identifier=None),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_torch(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a proxy for an internal PyTorch type
        """

        class FakeTorch(object):
            """Not a real torch"""

            def __str__(self):
                """But a real str

                :returns: An actual str
                :rtype: ```Literal['<required parameter>']```
                """
                return "<required parameter>"

        # type("FakeTorch", tuple(), {"__str__": lambda _: "<required parameter>"})

        run_ast_test(
            gen_ast=param2argparse_param(
                (
                    "byo",
                    {
                        "default": FakeTorch(),
                    },
                ),
            ),
            gold=Expr(
                Call(
                    args=[set_value("--byo")],
                    func=Attribute(
                        Name("argument_parser", Load()),
                        "add_argument",
                        Load(),
                    ),
                    keywords=[
                        keyword(
                            arg="type",
                            value=Name(FakeTorch.__name__, Load()),
                            identifier=None,
                        ),
                        keyword(arg="required", value=set_value(True), identifier=None),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            test_case_instance=self,
        )

    def test_param2argparse_param_default_class_str(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a proxy for an unexpected type
        """

        self.assertEqual(
            get_value(
                param2argparse_param(
                    (
                        "byo",
                        {"default": 5.5, "typ": "<class 'float'>"},
                    ),
                )
                .value.keywords[0]
                .value
            ),
            "float",
        )

    def test_param2argparse_param_default_notimplemented(self) -> None:
        """
        Tests that param2argparse_param works to change the type based on the default
          whence said default is a proxy for an unexpected type
        """

        with self.assertRaises(NotImplementedError) as cm:
            param2argparse_param(
                (
                    "byo",
                    {
                        "default": memoryview(b""),
                    },
                ),
            )
        self.assertEqual(
            *map(
                lambda s: s[: s.rfind("<") + 10],
                (
                    "Parsing type <class 'memoryview'>, which contains <memory at 0x10bb8c400>",
                    str(cm.exception),
                ),
            )
        )

    def test_parse_to_scalar(self) -> None:
        """Test various inputs and outputs for `parse_to_scalar`"""
        for fst, snd in (
            (5, 5),
            ("5", "5"),
            (set_value(5), 5),
            (ast.Expr(None), NoneStr),
        ):
            self.assertEqual(parse_to_scalar(fst), snd)

        self.assertEqual(
            get_value(parse_to_scalar(ast.parse("[5]").body[0]).elts[0]), 5
        )
        self.assertTrue(
            cmp_ast(
                parse_to_scalar(ast.parse("[5]").body[0]),
                List([set_value(5)], Load()),
            )
        )

        self.assertEqual(parse_to_scalar(ast.parse("[5]")), "[5]")

        parse_to_scalar(ast.parse("[5]").body[0])

        self.assertRaises(NotImplementedError, parse_to_scalar, memoryview(b""))
        self.assertRaises(NotImplementedError, parse_to_scalar, memoryview(b""))

    def test_infer_type_and_default(self) -> None:
        """Test edge cases for `infer_type_and_default`"""
        self.assertTupleEqual(
            infer_type_and_default(None, 5, "str", False), (None, 5, False, "int")
        )

        self.assertTupleEqual(
            infer_type_and_default(None, [5], "str", False), ("append", 5, False, "int")
        )

        self.assertTupleEqual(
            infer_type_and_default(None, tuple(range(5)), "str", False),
            (None, str(list(range(5))), False, "loads"),
        )

        self.assertTupleEqual(
            infer_type_and_default(None, 0j, "str", False), (None, 0j, False, "complex")
        )

        self.assertTupleEqual(
            infer_type_and_default(None, None, "str", False), (None, None, False, None)
        )

        self.assertTupleEqual(
            infer_type_and_default(None, "```{(5,6): (7,8)}```", "str", False),
            (None, "- 5\n- 6\n", False, "loads"),
        )

    def test__parse_default_from_ast(self) -> None:
        """
        Test `_parse_default_from_ast`
        """
        self.assertTupleEqual(
            _parse_default_from_ast(None, ast.parse("[1,2,56]").body[0], True, None),
            (None, "[1, 2, 56]", True, "loads"),
        )
        self.assertTupleEqual(
            _parse_default_from_ast(None, ast.parse("[5]").body[0], True, None),
            ("append", 5, True, "int"),
        )


unittest_main()



"""
Given the truth, show others the path
"""

from ast import ClassDef, FunctionDef, Module
from collections import OrderedDict
from os import path

from cdd import emit, parse
from cdd.ast_utils import RewriteAtQuery, cmp_ast, find_in_ast, get_function_type
from cdd.pure_utils import pluralise, strip_split
from cdd.source_transformer import ast_parse


def _default_options(node, search, type_wanted):
    """
    Conform the given file to the `intermediate_repr`

    :param node: AST node
    :type node: ```AST```

    :param search: Search query, e.g., ['node_name', 'function_name', 'arg_name']
    :type search: ```List[str]```

    :param type_wanted: AST instance
    :type type_wanted: ```AST```

    :returns: Arguments to pass to `emit.*` function
    :rtype: ```Callable[[], dict]```
    """
    return {
        "FunctionDef": lambda: {
            "function_type": node if node is None else get_function_type(node),
            "function_name": search[-1] if len(search) else "set_cli_args",
        },
        "ClassDef": lambda: {
            "class_name": search[-1] if len(search) else "ConfigClass",
        },
    }.get(type_wanted.__name__, lambda: {})


def _get_name_from_namespace(args, fun_name):
    """
    Gets the arg from the namespace which matches the given prefix

    :param args: Namespace with the values of the CLI arguments
    :type args: ```Namespace```

    :param fun_name: Name of the start of the function
    :type fun_name: ```str```

    :returns: Argument from Namespace object
    :rtype: ```str```
    """
    return next(
        getattr(args, pluralise(arg))[0]
        for arg in vars(args).keys()
        if arg == "_".join((fun_name, "names"))
    )


def ground_truth(args, truth_file):
    """
    There is but one truth. Conform.

    :param args: Namespace with the values of the CLI arguments
    :type args: ```Namespace```

    :param truth_file: contains the filename of the one true source
    :type truth_file: ```str```

    :returns: Filenames and whether they were changed
    :rtype: ```OrderedDict```
    """
    arg2parse_emit_type = {
        "argparse_function": (parse.argparse_ast, emit.argparse_function, FunctionDef),
        "class": (parse.class_, emit.class_, ClassDef),
        "function": (parse.function, emit.function, FunctionDef),
    }

    parse_func, emit_func, type_wanted = arg2parse_emit_type[args.truth]
    search = _get_name_from_namespace(args, args.truth).split(".")

    with open(truth_file, "rt") as f:
        true_ast = ast_parse(f.read(), filename=truth_file)

    original_node = find_in_ast(search, true_ast)
    gold_ir = parse_func(
        original_node,
        **_default_options(node=original_node, search=search, type_wanted=type_wanted)()
    )

    effect = OrderedDict()
    # filter(lambda arg: arg != args.truth, arg2parse_emit_type.keys()):
    for fun_name, (parse_func, emit_func, type_wanted) in arg2parse_emit_type.items():
        search = list(strip_split(_get_name_from_namespace(args, fun_name), "."))

        filenames = getattr(args, pluralise(fun_name))
        assert isinstance(
            filenames, (list, tuple)
        ), "Expected Union[list, tuple] got {!r}".format(type(filenames).__name__)

        effect.update(
            map(
                lambda filename: _conform_filename(
                    filename=filename,
                    search=search,
                    emit_func=emit_func,
                    replacement_node_ir=gold_ir,
                    type_wanted=type_wanted,
                ),
                filenames,
            )
        )

    return effect


def _conform_filename(
    filename,
    search,
    emit_func,
    replacement_node_ir,
    type_wanted,
):
    """
    Conform the given file to the `intermediate_repr`

    :param filename: Location of file
    :type filename: ```str```

    :param search: Search query, e.g., ['node_name', 'function_name', 'arg_name']
    :type search: ```List[str]```

    :param replacement_node_ir: Replace what is found with the contents of this param
    :type replacement_node_ir: ```dict```

    :param type_wanted: AST instance
    :type type_wanted: ```AST```

    :returns: filename, whether the file was modified
    :rtype: ```Tuple[str, bool]```
    """
    filename = path.realpath(path.expanduser(filename))

    if not path.isfile(filename):
        emit.file(
            emit_func(
                replacement_node_ir,
                emit_default_doc=False,  # emit_func.__name__ == "class_"
            ),
            filename=filename,
            mode="wt",
            skip_black=False,
        )
        return filename, True

    with open(filename, "rt") as f:
        parsed_ast = ast_parse(f.read(), filename=filename)
    assert isinstance(parsed_ast, Module)

    original_node = find_in_ast(search, parsed_ast)
    replacement_node = emit_func(
        replacement_node_ir,
        **_default_options(node=original_node, search=search, type_wanted=type_wanted)()
    )
    if original_node is None:
        emit.file(replacement_node, filename=filename, mode="a", skip_black=False)
        return filename, True
    assert len(search) > 0

    assert type(replacement_node) == type_wanted, "Expected {!r} got {!r}".format(
        type_wanted, type(replacement_node).__name__
    )

    replaced = False
    if not cmp_ast(original_node, replacement_node):
        rewrite_at_query = RewriteAtQuery(
            search=search,
            replacement_node=replacement_node,
        )
        rewrite_at_query.visit(parsed_ast)

        print(
            "modified" if rewrite_at_query.replaced else "unchanged", filename, sep="\t"
        )
        if rewrite_at_query.replaced:
            emit.file(parsed_ast, filename, mode="wt", skip_black=False)

        replaced = rewrite_at_query.replaced

    return filename, replaced


__all__ = ["ground_truth"]



"""
Helpers to traverse the AST, extract the docstring out, parse and format to intended style
"""
from ast import (
    AnnAssign,
    Assign,
    AsyncFunctionDef,
    ClassDef,
    FunctionDef,
    NodeTransformer,
    get_docstring,
    walk,
)
from collections import OrderedDict
from operator import attrgetter

from cdd import emit, parse
from cdd.ast_utils import find_in_ast, set_arg, set_docstring, to_annotation
from cdd.docstring_parsers import parse_docstring
from cdd.parser_utils import ir_merge


def has_type_annotations(node):
    """
    Whether the node—incl. any nodes within this node—have type annotations

    :param node: AST node
    :type node: ```AST```
    """
    return any(
        filter(
            lambda _node: hasattr(_node, "annotation")
            and _node.annotation is not None
            or hasattr(_node, "returns")
            and _node.returns is not None,
            walk(node),
        )
    )


class DocTrans(NodeTransformer):
    """
    Walk the nodes modifying the docstring and inlining||commenting types as it goes
    """

    def __init__(
        self, docstring_format, type_annotations, existing_type_annotations, whole_ast
    ):
        """
        Transform the docstrings found to intended docstring_format, potentially manipulating type annotations also

        :param docstring_format: Format of docstring
        :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

        :param type_annotations: True to have type annotations (3.6+), False to place in docstring
        :type type_annotations: ```bool```

        :param existing_type_annotations: Whether there are any type annotations (3.6+)
        :type existing_type_annotations: ```bool```

        :param whole_ast: The entire input AST, useful for lookups by location
        :type whole_ast: ```AST``
        """
        self.docstring_format = docstring_format
        self.type_annotations = type_annotations
        self.existing_type_annotations = existing_type_annotations
        self.whole_ast = whole_ast
        self.memoized = {}

    # def generic_visit(self, node):
    #     """
    #     visits the `AST` node, if it could have a docstring pass it off to that handler
    #
    #     :param node: The AST node
    #     :type node: ```AST```
    #
    #     :returns: Potentially changed AST node
    #     :rtype: ```AST```
    #     """
    #     is_func, doc_str = isinstance(node, (AsyncFunctionDef, FunctionDef)), None
    #     if is_func or isinstance(node, ClassDef):
    #         node, doc_str = self._handle_node_with_docstring(node)
    #         doc_str = ast.get_docstring(node)
    #     if is_func:
    #         node = self._handle_function(node, doc_str)
    #     return super(DocTrans, self).generic_visit(node)

    def visit_AnnAssign(self, node):
        """
        Handle `AnnAssign`

        :param node: AnnAssign
        :type node: ```AnnAssign```

        :returns: `AnnAssign` if `type_annotations` and type found else `Assign`
        :rtype: ```Union[AnnAssign, Assign]```
        """
        if self.type_annotations:
            node.annotation = self._get_ass_typ(node)
            node.type_comment = None
            return node

        return Assign(
            targets=[node.target],
            value=node.value,
            type_comment=node.annotation,
            lineno=None,
        )

    def visit_Assign(self, node):
        """
        Handle `Assign`

        :param node: Assign
        :type node: ```Assign```

        :returns: `AnnAssign` if `type_annotations` and type found else `Assign`
        :rtype: ```Union[Assign, AnnAssign]```
        """
        typ = self._get_ass_typ(node)
        if self.type_annotations:
            assert len(node.targets) == 1
            return AnnAssign(
                annotation=typ,
                value=node.value,
                type_comment=None,
                lineno=None,
                simple=1,
                target=node.targets[0],
                expr=None,
                expr_target=None,
                expr_annotation=None,
            )
        else:
            node.type_comment = typ
        return node

    def visit_Module(self, node):
        """
        visits the `Module`, potentially augmenting its docstring indentation

        :param node: Module
        :type node: ```Module```

        :returns: Potentially changed Module
        :rtype: ```Module```
        """
        # Clean might be wrong if the header is a license or other long-spiel documentation
        doc_str = get_docstring(node, clean=True)
        empty = doc_str is None
        if not empty:
            set_docstring("\n{}\n".format(doc_str), empty, node)
        node.body = list(map(self.visit, node.body))
        return node

    def visit_FunctionDef(self, node):
        """
        visits the `FunctionDef`, potentially augmenting its docstring and argument types

        :param node: FunctionDef
        :type node: ```FunctionDef```

        :returns: Potentially changed FunctionDef
        :rtype: ```FunctionDef```
        """
        return self._handle_function(node, get_docstring(node))

    def _get_ass_typ(self, node):
        """
        Get the type of the assignment

        :param node: Assignment
        :type node: ```Union[Assign, AnnAssign]```

        :returns: The type of the assignment, e.g., `int`
        :rtype: ```Optional[str]```
        """
        name, typ_dict = (
            (node.targets[0], {"typ": node.type_comment})
            if isinstance(node, Assign)
            else (node.target, {"typ": node.annotation or node.type_comment})
        )
        if not hasattr(node, "_location"):
            return typ_dict["typ"]
        search = node._location[:-1]
        search_str = ".".join(search)

        self.memoized[search_str] = ir = (
            self.memoized.get(
                search_str,
                (
                    lambda parent: (
                        (
                            lambda doc_str: None
                            if doc_str is None
                            else parse.docstring(doc_str)
                        )(get_docstring(parent))
                        if isinstance(parent, (ClassDef, AsyncFunctionDef, FunctionDef))
                        else {"params": OrderedDict()}
                    )
                )(find_in_ast(search, self.whole_ast)),
            )
            or {"params": OrderedDict()}
        )

        return ir["params"].get(name, typ_dict)[
            "typ"
        ]  # if ir is not None and ir.get("params") else None

    def _handle_function(self, node, doc_str):
        """
        Handle functions

        :param node: AsyncFunctionDef | FunctionDef
        :type node: ```Union[AsyncFunctionDef, FunctionDef]```

        :param doc_str: The docstring
        :type doc_str: ```Optional[str]```

        :returns: Same type as input with args, returns, and docstring potentially modified
        :rtype: ```Union[AsyncFunctionDef, FunctionDef]```
        """
        ir = parse_docstring(doc_str)
        ir_merge(ir, parse.function(node))
        ir["name"] = node.name
        indent_level = max(
            len(node._location) - 1, 1
        )  # function docstrings always have at least 1 indent level
        _doc_str = emit.docstring(
            ir,
            emit_types=not self.type_annotations,
            emit_default_doc=False,
            docstring_format=self.docstring_format,
            indent_level=indent_level,
        )
        if _doc_str.isspace():
            if doc_str is not None:
                del node.body[0]
        else:
            set_docstring(_doc_str, False, node)
        if self.type_annotations:
            # Add annotations
            if ir["params"]:
                node.args.args = list(
                    map(
                        lambda _arg: set_arg(
                            _arg.arg,
                            annotation=to_annotation(
                                _arg.annotation
                                if _arg.annotation is not None
                                else ir["params"][_arg.arg].get("typ")
                            ),
                        ),
                        node.args.args,
                    )
                )
            if (
                "return_type" in (ir.get("returns") or iter(()))
                and ir["returns"]["return_type"].get("typ") is not None
            ):
                node.returns = to_annotation(ir["returns"]["return_type"]["typ"])
        else:
            # Remove annotations
            node.args.args = list(map(set_arg, map(attrgetter("arg"), node.args.args)))
            node.returns = None

        node.body = list(map(self.visit, node.body))
        return node


def clear_annotation(node):
    """
    Remove annotations and type_comments from node

    :param node: AST node
    :type node: ```AST```

    :returns: AST node with annotations and type_comments set to `None`
    :rtype: ```AST```
    """
    if getattr(node, "annotation", None) is not None:
        node.annotation = None
    if getattr(node, "type_comment", None) is not None:
        node.type_comment = None
    return node


__all__ = ["DocTrans", "has_type_annotations"]



"""
Tests for reeducation
"""
import os
from argparse import Namespace
from ast import FunctionDef
from copy import deepcopy
from functools import partial
from io import StringIO
from os import path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cdd import emit
from cdd.conformance import _conform_filename, _get_name_from_namespace, ground_truth
from cdd.tests.mocks.argparse import argparse_func_ast
from cdd.tests.mocks.classes import class_ast_no_default_doc
from cdd.tests.mocks.ir import intermediate_repr
from cdd.tests.mocks.methods import class_with_method_ast, class_with_method_types_ast
from cdd.tests.utils_for_tests import unittest_main

"""
# type: Final[bool]
"""
modified = False


class TestConformance(TestCase):
    """
    Tests must comply. They shall be assimilated.
    """

    def test_ground_truth(self) -> None:
        """Straight from the ministry. Absolutely."""

        with TemporaryDirectory() as tempdir:
            self.assertTupleEqual(
                tuple(
                    map(
                        lambda filename_unmodified: (
                            os.path.basename(filename_unmodified[0]),
                            filename_unmodified[1],
                        ),
                        self.ground_truth_tester(
                            tempdir=tempdir,
                        )[0].items(),
                    )
                ),
                (("argparse.py", False), ("classes.py", False), ("methods.py", False)),
            )

    def test_ground_truths(self) -> None:
        """My truth is being tested."""

        with TemporaryDirectory() as tempdir:
            tempdir_join = partial(path.join, tempdir)

            argparse_functions = list(
                map(
                    lambda i: (
                        lambda argparse_function: emit.file(
                            argparse_func_ast, argparse_function, mode="wt"
                        )
                        or argparse_function
                    )(tempdir_join("argparse{i}.py".format(i=i))),
                    range(10),
                )
            )
            # Test if can create missing file
            argparse_functions.append(tempdir_join("argparse_missing.py"))

            # Test if can fill in empty file
            argparse_functions.append(tempdir_join("argparse_empty.py"))

            class_ = tempdir_join("classes.py")
            emit.file(class_ast_no_default_doc, class_, mode="wt")

            function = tempdir_join("methods.py")
            emit.file(class_with_method_ast, function, mode="wt")

            args = Namespace(
                **{
                    "argparse_functions": argparse_functions,
                    "argparse_function_names": ("set_cli_args",),
                    "classes": (class_,),
                    "class_names": ("ConfigClass",),
                    "functions": (function,),
                    "function_names": ("C.function_name",),
                    "truth": "argparse_function",
                }
            )
            with patch("sys.stdout", new_callable=StringIO), patch(
                "sys.stderr", new_callable=StringIO
            ):
                res = ground_truth(
                    args,
                    argparse_functions[0],
                )

            self.assertTupleEqual(
                tuple(
                    map(
                        lambda filename_unmodified: (
                            os.path.basename(filename_unmodified[0]),
                            filename_unmodified[1],
                        ),
                        res.items(),
                    )
                ),
                (
                    ("argparse0.py", False),
                    ("argparse1.py", False),
                    ("argparse2.py", False),
                    ("argparse3.py", False),
                    ("argparse4.py", False),
                    ("argparse5.py", False),
                    ("argparse6.py", False),
                    ("argparse7.py", False),
                    ("argparse8.py", False),
                    ("argparse9.py", False),
                    ("argparse_missing.py", True),
                    ("argparse_empty.py", True),
                    ("classes.py", False),
                    ("methods.py", False),
                ),
            )

    def test_ground_truth_changes(self) -> None:
        """Time for a new master."""

        ir = deepcopy(intermediate_repr)
        ir["returns"]["return_type"]["typ"] = "Tuple[np.ndarray, np.ndarray]"

        with TemporaryDirectory() as tempdir:
            self.assertTupleEqual(
                tuple(
                    map(
                        lambda filename_unmodified: (
                            os.path.basename(filename_unmodified[0]),
                            filename_unmodified[1],
                        ),
                        self.ground_truth_tester(
                            tempdir=tempdir,
                            _class_ast=emit.class_(ir, emit_default_doc=False),
                        )[0].items(),
                    )
                ),
                (("argparse.py", False), ("classes.py", True), ("methods.py", False)),
            )

    @staticmethod
    def ground_truth_tester(
        tempdir,
        _argparse_func_ast=argparse_func_ast,
        _class_ast=class_ast_no_default_doc,
        _class_with_method_ast=class_with_method_types_ast,
    ):
        """
        Helper for ground_truth tests

        :param tempdir: temporary directory
        :type tempdir: ```str```

        :param _argparse_func_ast: AST node
        :type _argparse_func_ast: ```FunctionDef```

        :param _class_ast: AST node
        :type _class_ast: ```ClassDef```

        :param _class_with_method_ast: AST node
        :type _class_with_method_ast: ```ClassDef```

        :returns: OrderedDict of filenames and whether they were changed, Args
        :rtype: ```Tuple[OrderedDict, Namespace]```
        """
        argparse_function = os.path.join(tempdir, "argparse.py")
        emit.file(_argparse_func_ast, argparse_function, mode="wt")

        class_ = os.path.join(tempdir, "classes.py")
        emit.file(_class_ast, class_, mode="wt")

        function = os.path.join(tempdir, "methods.py")
        emit.file(_class_with_method_ast, function, mode="wt")

        args = Namespace(
            **{
                "argparse_functions": (argparse_function,),
                "argparse_function_names": ("set_cli_args",),
                "classes": (class_,),
                "class_names": ("ConfigClass",),
                "functions": (function,),
                "function_names": ("C.function_name",),
                "truth": "argparse_function",
            }
        )

        with patch("sys.stdout", new_callable=StringIO), patch(
            "sys.stderr", new_callable=StringIO
        ):
            return (
                ground_truth(
                    args,
                    argparse_function,
                ),
                args,
            )

    def test__get_name_from_namespace(self) -> None:
        """Test `_get_name_from_namespace`"""
        args = Namespace(foo_names=("bar",))
        self.assertEqual(_get_name_from_namespace(args, "foo"), args.foo_names[0])

    def test__conform_filename_nonexistent(self) -> None:
        """Tests that _conform_filename returns the right result"""

        with TemporaryDirectory() as tempdir:
            argparse_function_filename = os.path.realpath(
                os.path.join(tempdir, "no_file_here.py")
            )

            self.assertTupleEqual(
                _conform_filename(
                    filename=argparse_function_filename,
                    search=["set_cli_args"],
                    emit_func=emit.argparse_function,
                    replacement_node_ir=deepcopy(intermediate_repr),
                    type_wanted=FunctionDef,
                ),
                (argparse_function_filename, True),
            )

    def test__conform_filename_filled(self) -> None:
        """Tests that _conform_filename returns the right result"""

        with TemporaryDirectory() as tempdir:
            argparse_function_filename = os.path.realpath(
                os.path.join(tempdir, "correct_contents.py")
            )

            emit.file(
                argparse_func_ast,
                argparse_function_filename,
                mode="wt",
            )

            self.assertTupleEqual(
                _conform_filename(
                    filename=argparse_function_filename,
                    search=["impossibru"],
                    emit_func=emit.argparse_function,
                    replacement_node_ir=deepcopy(intermediate_repr),
                    type_wanted=FunctionDef,
                ),
                (argparse_function_filename, True),
            )

    def test__conform_filename_unchanged(self) -> None:
        """Tests that _conform_filename returns the right result"""

        with TemporaryDirectory() as tempdir:
            argparse_function_filename = os.path.realpath(
                os.path.join(tempdir, "do_not_touch_this.py")
            )

            emit.file(argparse_func_ast, argparse_function_filename, mode="wt")
            with patch("sys.stdout", new_callable=StringIO):
                self.assertTupleEqual(
                    _conform_filename(
                        filename=argparse_function_filename,
                        search=["set_cli_args"],
                        emit_func=emit.argparse_function,
                        replacement_node_ir=deepcopy(intermediate_repr),
                        type_wanted=FunctionDef,
                    ),
                    (argparse_function_filename, False),
                )


unittest_main()



"""
Tests for source_transformer
"""
from ast import ClassDef
from unittest import TestCase
from unittest.mock import patch

from cdd.pure_utils import PY_GTE_3_9
from cdd.tests.utils_for_tests import unittest_main


class TestSourceTransformer(TestCase):
    """
    Tests for source_transformer
    """

    def test_to_code(self) -> None:
        """
        Tests to_source in Python 3.9 and < 3.9
        """
        class_def = ClassDef(
            name="Classy",
            bases=tuple(),
            decorator_list=[],
            body=[],
            keywords=tuple(),
            identifier_name=None,
            expr=None,
        )

        with patch("cdd.source_transformer.version_info", (3, 9, 0)):
            import cdd.source_transformer

            self.assertEqual(
                cdd.source_transformer.to_code(class_def).rstrip("\n"),
                "class Classy:",
            ) if PY_GTE_3_9 else self.assertRaises(
                AttributeError, lambda: cdd.source_transformer.to_code(class_def)
            )

        with patch("cdd.source_transformer.version_info", (3, 8, 0)):
            import cdd.source_transformer

            self.assertEqual(
                "class Classy:"
                if PY_GTE_3_9
                else cdd.source_transformer.to_code(class_def).rstrip("\n"),
                "class Classy:",
            )


unittest_main()



""" Tests for CLI sync_properties subparser (__main__.py) """
import os
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, run_cli_test, unittest_main


class TestCliSyncProperties(TestCase):
    """Test class for __main__.py"""

    def test_sync_properties_fails(self) -> None:
        """Tests CLI interface failure cases"""
        run_cli_test(
            self,
            ["sync_properties", "--wrong"],
            exit_code=2,
            output="the following arguments are required:"
            " --input-filename, --input-param, --output-filename, --output-param\n",
        )

    def test_non_existent_file_fails(self) -> None:
        """Tests nonexistent file throws the right error"""
        with TemporaryDirectory() as tempdir:
            filename = os.path.join(
                tempdir,
                "delete_this_1{}".format(os.path.basename(__file__)),
            )

            run_cli_test(
                self,
                [
                    "sync_properties",
                    "--input-file",
                    filename,
                    "--input-param",
                    "Foo.g.f",
                    "--output-file",
                    filename,
                    "--output-param",
                    "f.h",
                ],
                exit_code=2,
                output="--input-file must be an existent file. Got: {!r}\n".format(
                    filename
                ),
            )

        with TemporaryDirectory() as tempdir:
            input_filename = os.path.join(
                tempdir,
                "input_filename.py",
            )
            output_filename = os.path.join(
                tempdir,
                "output_filename.py",
            )
            open(input_filename, "wt").close()

            run_cli_test(
                self,
                [
                    "sync_properties",
                    "--input-file",
                    input_filename,
                    "--input-param",
                    "Foo.g.f",
                    "--output-file",
                    output_filename,
                    "--output-param",
                    "f.h",
                ],
                exit_code=2,
                output="--output-file must be an existent file. Got: {!r}\n".format(
                    output_filename
                ),
            )

    def test_sync_properties(self) -> None:
        """Tests CLI interface gets all the way to the sync_properties call without error"""
        with TemporaryDirectory() as tempdir:
            input_filename = os.path.join(tempdir, "class_.py")
            output_filename = os.path.join(tempdir, "method.py")
            open(input_filename, "wt").close()
            open(output_filename, "wt").close()

            with patch("cdd.__main__.sync_properties", mock_function):
                self.assertTrue(
                    run_cli_test(
                        self,
                        [
                            "sync_properties",
                            "--input-file",
                            input_filename,
                            "--input-param",
                            "Foo.g.f",
                            "--output-file",
                            output_filename,
                            "--output-param",
                            "f.h",
                        ],
                        exit_code=None,
                        output=None,
                    ),
                )


unittest_main()



"""
Generate routes
"""
import ast
from ast import Attribute, Call, ClassDef, FunctionDef, Name
from itertools import chain
from operator import attrgetter, itemgetter
from os import path

from cdd import parse
from cdd.ast_utils import get_value
from cdd.pure_utils import filename_from_mod_or_filename, rpartial
from cdd.routes import emit as routes_emit
from cdd.routes.parse import methods
from cdd.source_transformer import to_code
from cdd.tests.mocks.routes import route_prelude


def gen_routes(app, model_path, model_name, crud, route):
    """
    Generate route(s)

    :param app: Variable name (Bottle App)
    :type app: ```str```

    :param model_path: The path/module-resolution whence the model is
    :type model_path: ```str```

    :param model_name: Name of the model to recover from the `model_path`
    :type model_name: ```str```

    :param crud: (C)reate (R)ead (U)pdate (D)elete, like "CRUD" for all or "CD" for "Create" and "Delete"
    :type crud: ```Union[Literal['C', 'R'], Literal['C', 'U'], Literal['C', 'D'], Literal['R', 'C'],
                         Literal['R', 'U'], Literal['R', 'D'], Literal['U', 'C'], Literal['U', 'R'],
                         Literal['U', 'D'], Literal['D', 'C'], Literal['D', 'R'], Literal['D', 'U'],
                         Literal['C', 'R', 'U'], Literal['C', 'R', 'D'], Literal['C', 'U', 'R'],
                         Literal['C', 'U', 'D'], Literal['C', 'D', 'R'], Literal['C', 'D', 'U'],
                         Literal['R', 'C', 'U'], Literal['R', 'C', 'D'], Literal['R', 'U', 'C'],
                         Literal['R', 'U', 'D'], Literal['R', 'D', 'C'], Literal['R', 'D', 'U'],
                         Literal['U', 'C', 'R'], Literal['U', 'C', 'D'], Literal['U', 'R', 'C'],
                         Literal['U', 'R', 'D'], Literal['U', 'D', 'C'], Literal['U', 'D', 'R'],
                         Literal['D', 'C', 'R'], Literal['D', 'C', 'U'], Literal['D', 'R', 'C'],
                         Literal['D', 'R', 'U'], Literal['D', 'U', 'C'], Literal['D', 'U', 'R']]```

    :param route: The path of the resource
    :type route: ```str```

    :returns: Iterator of functions representing relevant CRUD operations
    :rtype: ```Iterator[FunctionDef]```
    """
    model_path = filename_from_mod_or_filename(model_path)

    assert path.isfile(model_path)
    with open(model_path, "rt") as f:
        mod = ast.parse(f.read())

    sqlalchemy_node = next(
        filter(
            lambda node: isinstance(node, ClassDef)
            and node.name == model_name
            or isinstance(node, Name)
            and node.id == model_name,
            ast.walk(mod),
        ),
        None,
    )
    sqlalchemy_ir = parse.sqlalchemy(sqlalchemy_node)
    primary_key = next(
        map(
            itemgetter(0),
            filter(
                lambda param: param[1]["doc"].startswith("[PK]"),
                sqlalchemy_ir["params"].items(),
            ),
        ),
        next(iter(sqlalchemy_ir["params"].keys())),
    )
    _route_config = {"app": app, "name": model_name, "route": route, "variant": -1}
    routes = []
    if "C" in crud:
        routes.append(routes_emit.create(**_route_config))
    _route_config["primary_key"] = primary_key

    funcs = {"R": routes_emit.read, "U": None, "D": routes_emit.destroy}
    routes.extend(funcs[key](**_route_config) for key in funcs if key in crud)
    return (
        map(itemgetter(0), map(attrgetter("body"), map(ast.parse, routes))),
        primary_key,
    )


def upsert_routes(app, routes, routes_path, route, primary_key):
    """
    Upsert the `routes` to the `routes_path`, on merge use existing body and replace interface/prototype

    :param app: Variable name (Bottle App)
    :type app: ```str```

    :param routes: Iterator of functions representing relevant CRUD operations
    :type routes: ```Iterator[FunctionDef]```

    :param route: The path of the resource
    :type route: ```str```

    :param primary_key: The primary key or id to lookup on for the route
    :type primary_key: ```str```

    :param routes_path: The path/module-resolution whence the routes are / will be
    :type routes_path: ```str```
    """
    routes_path = filename_from_mod_or_filename(routes_path)

    if not path.isfile(routes_path):
        with open(routes_path, "wt") as f:
            f.write(
                "\n\n".join(
                    chain.from_iterable(
                        (
                            (
                                route_prelude.replace(
                                    "rest_api =", "{app} =".format(app=app)
                                ),
                            ),
                            map(to_code, routes),
                        )
                    )
                )
            )
        return

    with open(routes_path, "rt") as f:
        mod = ast.parse(f.read())

    def get_names(functions):
        """
        Derive a method_name -> FunctionDef dictionary

        :param functions: Routing functions
        :type functions: ```Iterator[FunctionDef]```

        :returns: Dict of `method_name` to `FunctionDef`
        :rtype: ```Dict[str, FunctionDef]```
        """
        return dict(
            map(
                lambda func: (
                    next(
                        map(
                            lambda call: call.func.attr,
                            filter(
                                lambda call: all(
                                    (
                                        isinstance(call.func, Attribute),
                                        call.func.attr in methods,
                                    )
                                ),
                                filter(rpartial(isinstance, Call), func.decorator_list),
                            ),
                        )
                    ),
                    func,
                ),
                functions,
            )
        )

    routes_required = get_names(routes)
    routes_existing = get_names(
        filter(
            lambda node: any(
                filter(
                    lambda call: all(
                        (
                            isinstance(call.func, Attribute),
                            call.func.attr in methods,
                            get_value(call.args[0])
                            == "{route}{rest}".format(
                                route=route,
                                rest=""
                                if call.func.attr == "post"
                                else "/:{primary_key}".format(primary_key=primary_key),
                            ),
                            call.func.value.id == app,
                        )
                    ),
                    filter(rpartial(isinstance, Call), node.decorator_list),
                )
            ),
            filter(rpartial(isinstance, FunctionDef), ast.walk(mod)),
        )
    )
    missing_routes = (
        routes_required.keys() & routes_existing.keys() ^ routes_required.keys()
    )

    if not missing_routes:
        return

    with open(routes_path, "a") as f:
        f.write(
            "\n\n".join(
                map(
                    to_code,
                    map(
                        routes_required.__getitem__,
                        sorted(
                            missing_routes,
                            key={
                                "post": 0,
                                "get": 1,
                                "update": 2,
                                "delete": 3,
                            }.__getitem__,
                        ),
                    ),
                )
            )
        )


__all__ = ["gen_routes", "upsert_routes"]



""" eval.py for testing with `sync_properties` """

import cdd.tests.mocks

_attr_within = frozenset(("mocks",))

get_modules = tuple(
    attr for attr in dir(cdd.tests) if not attr.startswith("_") and attr in _attr_within
)

__all__ = ["get_modules"]



"""
OpenAPI mocks
"""
from cdd.tests.mocks.json_schema import config_schema, server_error_schema
from cdd.tests.mocks.routes import route_config

openapi_dict = {
    "openapi": "3.0.0",
    "info": {"title": "REST API", "version": "0.0.1"},
    "components": {
        "requestBodies": {
            "ConfigBody": {
                "content": {
                    "application/json": {
                        "schema": {
                            "$ref": "#/components/schemas/{name}".format(
                                name=route_config["name"]
                            )
                        }
                    }
                },
                "description": "A `{name}` object.".format(name=route_config["name"]),
                "required": True,
            }
        },
        "schemas": {
            name: {k: v for k, v in schema.items() if not k.startswith("$")}
            for name, schema in {
                route_config["name"]: config_schema,
                "ServerError": server_error_schema,
            }.items()
        },
    },
    "paths": {
        route_config["route"]: {
            "post": {
                "requestBody": {
                    "$ref": "#/components/requestBodies/{name}Body".format(
                        name=route_config["name"]
                    ),
                    "required": True,
                },
                "responses": {
                    "201": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/{name}".format(
                                        name=route_config["name"]
                                    )
                                }
                            }
                        },
                        "description": "A `{name}` object.".format(
                            name=route_config["name"]
                        ),
                    },
                    "400": {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/ServerError"}
                            }
                        },
                        "description": "A `ServerError` object.",
                    },
                },
                "summary": "A `{name}` object.".format(name=route_config["name"]),
            }
        },
        "{route_config[route]}/{{{route_config[primary_key]}}}".format(
            route_config=route_config
        ): {
            "delete": {
                "responses": {"204": {}},
                "summary": "Delete one `{name}`".format(name=route_config["name"]),
            },
            "get": {
                "responses": {
                    "200": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/{name}".format(
                                        name=route_config["name"]
                                    )
                                }
                            }
                        },
                        "description": "A `{name}` object.".format(
                            name=route_config["name"]
                        ),
                    },
                    "404": {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/ServerError"}
                            }
                        },
                        "description": "A `ServerError` object.",
                    },
                },
                "summary": "A `{name}` object.".format(name=route_config["name"]),
            },
            "parameters": [
                {
                    "description": "Primary key of target `{name}`".format(
                        name=route_config["name"]
                    ),
                    "in": "path",
                    "name": route_config["primary_key"],
                    "required": True,
                    "schema": {"type": "string"},
                }
            ],
        },
    },
}

__all__ = ["openapi_dict"]



""" Tests for sync_properties """

import ast
import os
from sys import modules
from tempfile import TemporaryDirectory
from unittest import TestCase

from pkg_resources import resource_filename

from cdd.pure_utils import PY_GTE_3_8, tab
from cdd.sync_properties import sync_properties
from cdd.tests.mocks.eval import get_modules
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


def populate_files(tempdir, input_str=None, output_str=None):
    """
    Populate files in the tempdir

    :param tempdir: Temporary directory
    :type tempdir: ```str```

    :param input_str: Input string to write to the input_filename
    :type input_str: ```Optional[str]```

    :param output_str: Output string to write to the output_filename
    :type output_str: ```Optional[str]```

    :returns: input filename, input str, input_str filename
    :rtype: ```Tuple[str, str, str]```
    """
    input_filename = os.path.join(tempdir, "class_.py")
    output_filename = os.path.join(tempdir, "method.py")
    input_str = input_str or (
        "from {package} import Literal\n\n"
        "class Foo(object):\n"
        "{tab}def g(f: Literal['a']):\n"
        "{tab}{tab}pass".format(
            package="typing" if PY_GTE_3_8 else "typing_extensions", tab=tab
        )
    )
    output_str = output_str or (
        "from {package} import Literal\n\n"
        "def f(h: Literal['b']):"
        "{tab}{tab}pass".format(
            package="typing" if PY_GTE_3_8 else "typing_extensions", tab=tab
        )
    )
    with open(input_filename, "wt") as f:
        f.write(input_str)
    with open(output_filename, "wt") as f:
        f.write(output_str)
    return input_filename, input_str, output_filename, output_str


class TestSyncProperties(TestCase):
    """Test class for sync_properties.py"""

    def test_sync_properties(self) -> None:
        """Tests `sync_properties` with `call=False`"""

        with TemporaryDirectory() as tempdir:
            (
                input_filename,
                input_str,
                output_filename,
                output_str,
            ) = populate_files(tempdir)

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("Foo.g.f",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("f.h",),
                )
            )

            self.run_sync_properties_test(
                input_filename,
                input_str,
                output_filename,
                ast.parse(output_str.replace("h: Literal['b']", "f: Literal['a']")),
            )

    def test_sync_properties_eval(self) -> None:
        """Tests `sync_properties` with `call=True`"""

        with open(
            os.path.join(
                os.path.dirname(
                    os.path.dirname(
                        resource_filename(modules[__name__].__name__, "__init__.py")
                    )
                ),
                "tests",
                "mocks",
                "eval.py",
            ),
            "rt",
        ) as f:
            eval_mock_str = f.read()

        with TemporaryDirectory() as tempdir:
            (
                input_filename,
                input_str,
                output_filename,
                output_str,
            ) = populate_files(tempdir, input_str=eval_mock_str)

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("get_modules",),
                    input_eval=True,
                    output_filename=output_filename,
                    output_params=("f.h",),
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=eval_mock_str,
                gold=ast.parse(
                    output_str.replace("h: Literal['b']", "h: Literal['mocks']")
                ),
            )

    def test_sync_properties_output_param_wrap(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set"""

        with TemporaryDirectory() as tempdir:
            (
                input_filename,
                input_str,
                output_filename,
                output_str,
            ) = populate_files(tempdir)

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("Foo.g.f",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("f.h",),
                    output_param_wrap="Optional[List[Union[{output_param}, str]]]",
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=input_str,
                gold=ast.parse(
                    output_str.replace(
                        "h: Literal['b']", "f: Optional[List[Union[Literal['a'], str]]]"
                    )
                ),
            )

    def test_sync_properties_output_param_wrap_no_annotation(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set when type annotation isn't being replaced"""

        with TemporaryDirectory() as tempdir:
            (input_filename, input_str, output_filename, output_str,) = populate_files(
                tempdir,
                input_str=(
                    "from {package} import Literal\n\n"
                    "class Foo(object):\n"
                    "{tab}def g(f):\n"
                    "{tab}{tab}pass".format(
                        package="typing" if PY_GTE_3_8 else "typing_extensions", tab=tab
                    )
                ),
            )

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("Foo.g.f",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("f.h",),
                    output_param_wrap="Optional[List[Union[{output_param}, str]]]",
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=input_str,
                gold=ast.parse(output_str.replace("h: Literal['b']", "f")),
            )

    def test_sync_properties_output_param_wrap_no_type(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set when replacement_node has no type"""

        with TemporaryDirectory() as tempdir:
            (input_filename, input_str, output_filename, output_str,) = populate_files(
                tempdir,
                output_str=(
                    "from {package} import Literal\n\n"
                    "class Foo(object):\n"
                    "{tab}def f(h):\n"
                    "{tab}{tab}pass".format(
                        package="typing" if PY_GTE_3_8 else "typing_extensions", tab=tab
                    )
                ),
            )

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("Foo.g.f",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("Foo.f.h",),
                    output_param_wrap="Optional[List[Union[{output_param}, str]]]",
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=input_str,
                gold=ast.parse(
                    output_str.replace(
                        "(h)", "(f: Optional[List[Union[Literal['a'], str]]])"
                    )
                ),
            )

    def test_sync_properties_output_param_wrap_subscript(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set when replacement_node is subscript and !input_eval"""

        with TemporaryDirectory() as tempdir:
            (input_filename, input_str, output_filename, output_str,) = populate_files(
                tempdir,
                input_str="a = tuple(range(5))",
                output_str="def j(k):\n" "{tab}pass\n".format(tab=tab),
            )

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("a",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("j.k",),
                    output_param_wrap=None,
                )
            )

            with open(output_filename, "rt") as f:
                # Technically this produces an invalid AST, but we don't care… `input_eval is False`
                self.assertEqual(
                    f.read().rstrip(),
                    output_str.replace("(k)", "(a: tuple(range(5)))").rstrip(),
                )

    def test_sync_properties_output_param_wrap_subscript_eval0(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set when replacement_node is subscript"""

        with TemporaryDirectory() as tempdir:
            (input_filename, input_str, output_filename, output_str,) = populate_files(
                tempdir, input_str="a = tuple(range(5))", output_str="def j(k): pass"
            )

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("a",),
                    input_eval=True,
                    output_filename=output_filename,
                    output_params=("j.k",),
                    output_param_wrap=None,
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=input_str,
                gold=ast.parse(
                    output_str.replace("(k)", "(k: Literal[0, 1, 2, 3, 4])")
                ),
            )

    def test_sync_properties_output_param_wrap_subscript_eval1(self) -> None:
        """Tests `sync_properties` with `output_param_wrap` set when replacement_node is subscript"""

        with TemporaryDirectory() as tempdir:
            (input_filename, input_str, output_filename, output_str,) = populate_files(
                tempdir,
                input_str="import pip\n"
                "c = { attr: getattr(pip, attr)"
                "      for attr in dir(pip)"
                "      if attr in frozenset(('__version__', '__name__')) }\n"
                "a = tuple(sorted(c.keys()))\n",
                output_str="class C(object):\n"
                "{tab}def j(k: str):\n"
                "{tab}{tab}pass\n".format(tab=tab),
            )

            self.assertIsNone(
                sync_properties(
                    input_filename=input_filename,
                    input_params=("a",),
                    input_eval=True,
                    output_filename=output_filename,
                    output_params=("C.j.k",),
                    output_param_wrap=None,
                )
            )

            self.run_sync_properties_test(
                input_filename=input_filename,
                output_filename=output_filename,
                input_str=input_str,
                gold=ast.parse(
                    output_str.replace(
                        "(k: str)", "(k: Literal['__name__', '__version__'])"
                    )
                ),
            )

    def test_sync_properties_output_param_wrap_fails(self) -> None:
        """Tests `sync_properties` fails with `output_param_wrap` set when replacement_node is unknown"""

        with TemporaryDirectory() as tempdir:
            (
                input_filename,
                input_str,
                output_filename,
                output_str,
            ) = populate_files(tempdir, input_str="local = locals()")

            self.assertRaises(
                NotImplementedError,
                lambda: sync_properties(
                    input_filename=input_filename,
                    input_params=("local",),
                    input_eval=False,
                    output_filename=output_filename,
                    output_params=("f.h",),
                    output_param_wrap="Optional[List[Union[{output_param}, str]]]",
                ),
            )

    def test_sync_properties_eval_fails(self) -> None:
        """Tests `sync_properties` fails with `call=True` and dots"""
        with TemporaryDirectory() as tempdir:
            input_filename = os.path.join(tempdir, "input_.py")
            output_filename = os.path.join(tempdir, "input_str.py")

            open(input_filename, "wt").close()
            open(output_filename, "wt").close()

            self.assertRaises(
                NotImplementedError,
                lambda: sync_properties(
                    input_filename=input_filename,
                    input_params=("foo.bar",),
                    input_eval=True,
                    output_filename=output_filename,
                    output_params=("f.h",),
                ),
            )

    def test_eval(self) -> None:
        """Ensure mock returns the right result"""
        self.assertTupleEqual(get_modules, ("mocks",))

    def run_sync_properties_test(
        self, input_filename, input_str, output_filename, gold
    ):
        """
        Common test for the suite

        :param input_filename: Filename of input
        :type input_filename: ```str```

        :param input_str: Python source
        :type input_str: ```str```

        :param output_filename: Filename of output
        :type output_filename: ```str```

        :param gold: Gold standard AST
        :type gold: ```Union[ast.Module, ast.ClassDef, ast.FunctionDef]```
        """

        # Confirm that the input_filename is unedited
        with open(input_filename, "rt") as f:
            run_ast_test(self, gen_ast=ast.parse(input_str), gold=ast.parse(f.read()))

        # Confirm that the output_filename is edited correctly
        with open(output_filename, "rt") as f:
            run_ast_test(
                self,
                gen_ast=ast.parse(f.read()),
                gold=gold,
            )


unittest_main()



"""
Constant strings and tuples of strings which are to be interpolated in `emit.py`
"""
from string import Template

from cdd.pure_utils import indent_all_but_first

_create_route_desc = indent_all_but_first(
    '''"""Create `{name}`

```yml
responses:
  '201':
    description: A `{name}` object.
    content:
      application/json:
        schema:
          $ref: ```{name}```
  '400':
    description: A `ServerError` object.
    content:
      application/json:
        schema:
          $ref: ```ServerError```
```

:returns: Created ```{name}``` instance (as a dict), or an error
:rtype: ```dict```
"""'''
)
_create_helper_desc = "Create / handle-errors with Bottle and SQLalchemy"

create_route_variants = tuple(
    map(
        lambda s: Template(s).substitute(_create_route_desc=_create_route_desc),
        (
            """@{app}.post({route!r})
def create():
    $_create_route_desc
    try:
        config = {name}(**request.json)
    except TypeError as e:
        response.status = 400
        return {{"error": "ValidationError", "error_description": "\\n".join(e.args)}}

    try:
        with Session(engine) as session:
            session.add(config)
            session.commit()
            created = orm_to_dict(config)
    except DatabaseError as e:
        response.status = 400
        return {{"error": e.__class__.__name__, "error_code": e.code, "error_description": str(e.__cause__)}}

    return created
""",
            """@{app}.post({route!r})
def create():
    $_create_route_desc
    return create_helper0(request, response)({name})
""",
            """@{app}.post({route!r})
def create():
    $_create_route_desc
    code, body = create_helper1({name}, request.body)
    response.status = code
    return body
""",
        ),
    )
)

create_helper_variants = tuple(
    map(
        lambda s: Template(s).substitute(_create_helper_desc=_create_helper_desc),
        (
            '''def create_helper0(req, res):
    """
    $_create_helper_desc

    :param req: Bottle request
    :type req: ```bottle.request```

    :param res: Bottle response
    :type res: ```bottle.response```

    :returns: A function which actually does the work
    :rtype: ```Callable[[Base], dict]```
    """
    def _create_helper(orm_class):
        """
        $_create_helper_desc

        :param orm_class: An ORM class inheriting SQLalchemy declarative base class
        :type orm_class: ```Base```

        :returns: Created (as a dict) or error dict
        :rtype: ```dict```
        """
        try:
            orm_instance = orm_class(**req.json)
        except TypeError as e:
            res.status = 400
            return {{"error": "ValidationError", "error_description": "\\n".join(e.args)}}

        try:
            with Session(engine) as session:
                session.add(orm_instance)
                session.commit()
                created = orm_to_dict(orm_instance)
        except DatabaseError as e:
            res.status = 400
            return {{"error": e.__class__.__name__, "error_code": e.code, "error_description": str(e.__cause__)}}

        return created

    return _create_helper
''',
            '''def create_helper1(orm_class, body):
    """
    $_create_helper_desc

    :param orm_class: An ORM class inheriting SQLalchemy declarative base class
    :type orm_class: ```Base```

    :param body: Body of the instance to create
    :type body: ```dict```

    :returns: Status code, created (as a dict) or error dict
    :rtype: ```Tuple[int, dict]```
    """
    try:
        orm_instance = orm_class(**body)
    except TypeError as e:
        return 400, {{"error": "ValidationError", "error_description": "\\n".join(e.args)}}

    try:
        with Session(engine) as session:
            session.add(orm_instance)
            session.commit()
            created = orm_to_dict(orm_instance)
    except DatabaseError as e:
        return 400, {{"error": e.__class__.__name__, "error_code": e.code, "error_description": str(e.__cause__)}}

    return 201, created
''',
        ),
    )
)


read_route_variants = (
    '''
@{app}.get("{route}/:{id}")
def read({id}):
    """
    Find one `{name}` or error

    ```yml
    responses:
      '200':
        description: A `{name}` object.
        content:
          application/json:
            schema:
              $ref: ```{name}```
      '404':
        description: A `ServerError` object.
        content:
          application/json:
            schema:
              $ref: ```ServerError```
    ```

    :param {id}: The primary key of `{name}`
    :type {id}: ```str```

    :returns: Found `{name}` (as a dict) or error dict
    :rtype: ```dict```
    """
    with Session(engine) as session:
        config = session.execute(select({name}).filter_by({id}={id})).one_or_none()

    if config is None:
        response.status = 404
        return {{"error": "NotFound", "error_description": "{name} not found"}}

    return orm_to_dict(config[0])
''',
)

delete_route_variants = (
    '''
@{app}.delete("{route}/:{id}")
def destroy({id}):
    """
    Delete one `{name}`

    ```yml
    responses:
      '204':
    ```

    :param {id}: The primary key of `{name}`
    :type {id}: ```str```

    :returns: Found `{name}` (as a dict) or error dict
    :rtype: ```dict```
    """
    with Session(engine) as session:
        session.query({name}).filter({name}.{id} == {id}).delete(
            synchronize_session="evaluate"
        )

    response.status = 204
''',
)

__all__ = [
    "create_route_variants",
    "create_helper_variants",
    "read_route_variants",
    "delete_route_variants",
]



"""
Transform from string or AST representations of input, to AST, file, or str input_str.
"""
import ast
from ast import (
    Assign,
    Attribute,
    Call,
    ClassDef,
    Expr,
    FunctionDef,
    Load,
    Module,
    Name,
    Return,
    Store,
    Tuple,
    arguments,
    keyword,
)
from collections import OrderedDict
from functools import partial
from importlib import import_module
from itertools import chain
from operator import add
from sys import modules
from textwrap import indent

from cdd.ast_utils import (
    get_value,
    maybe_type_comment,
    param2argparse_param,
    param2ast,
    set_arg,
    set_value,
)
from cdd.docstring_utils import ARG_TOKENS, RETURN_TOKENS, emit_param_str
from cdd.emitter_utils import (
    RewriteName,
    _make_call_meth,
    generate_repr_method,
    get_internal_body,
    param2json_schema_property,
    param_to_sqlalchemy_column_call,
)
from cdd.pure_utils import (
    PY3_8,
    code_quoted,
    deindent,
    emit_separating_tabs,
    fill,
    identity,
    none_types,
    rpartial,
    simple_types,
    tab,
)
from cdd.source_transformer import to_code

black = (
    import_module("black")
    if "black" in modules
    else type(
        "black",
        tuple(),
        {
            "format_str": lambda src_contents, mode: src_contents,
            "Mode": lambda target_versions, line_length, is_pyi, string_normalization: None,
        },
    )
)


def argparse_function(
    intermediate_repr,
    emit_default_doc=False,
    function_name="set_cli_args",
    function_type="static",
    wrap_description=False,
    word_wrap=True,
    docstring_format="rest",
):
    """
    Convert to an argparse FunctionDef

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param function_name: name of function_def
    :type function_name: ```str```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Literal['self', 'cls', 'static']```

    :param wrap_description: Whether to word-wrap the description. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type wrap_description: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :returns:  AST node for function definition which constructs argparse
    :rtype: ```FunctionDef```
    """
    function_name = function_name or intermediate_repr["name"]
    function_type = function_type or intermediate_repr["type"]
    internal_body = get_internal_body(
        target_name=function_name,
        target_type=function_type,
        intermediate_repr=intermediate_repr,
    )

    return FunctionDef(
        args=arguments(
            args=[set_arg("argument_parser")],
            # None if function_type in frozenset((None, "static"))
            # else set_arg(function_type),
            defaults=[],
            kw_defaults=[],
            kwarg=None,
            kwonlyargs=[],
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=list(
            chain.from_iterable(
                (
                    iter(
                        (
                            Expr(
                                set_value(
                                    indent(
                                        docstring(
                                            {
                                                "doc": "Set CLI arguments",
                                                "params": OrderedDict(
                                                    (
                                                        (
                                                            "argument_parser",
                                                            {
                                                                "doc": "argument parser",
                                                                "typ": "ArgumentParser",
                                                            },
                                                        ),
                                                    )
                                                ),
                                                "returns": OrderedDict(
                                                    (
                                                        (
                                                            "return_type",
                                                            {
                                                                "doc": "argument_parser, {}".format(
                                                                    intermediate_repr[
                                                                        "returns"
                                                                    ]["return_type"][
                                                                        "doc"
                                                                    ]
                                                                )
                                                                if intermediate_repr[
                                                                    "returns"
                                                                ]["return_type"].get(
                                                                    "doc"
                                                                )
                                                                else "argument_parser",
                                                                "typ": "Tuple[ArgumentParser, {typ}]".format(
                                                                    typ=intermediate_repr[
                                                                        "returns"
                                                                    ][
                                                                        "return_type"
                                                                    ][
                                                                        "typ"
                                                                    ]
                                                                ),
                                                            }
                                                            if "return_type"
                                                            in (
                                                                (
                                                                    intermediate_repr
                                                                    or {}
                                                                ).get("returns")
                                                                or iter(())
                                                            )
                                                            and intermediate_repr[
                                                                "returns"
                                                            ]["return_type"].get("typ")
                                                            not in none_types
                                                            else {
                                                                "doc": "argument_parser",
                                                                "typ": "ArgumentParser",
                                                            },
                                                        ),
                                                    ),
                                                ),
                                            },
                                            docstring_format=docstring_format,
                                            word_wrap=word_wrap,
                                        ),
                                        tab,
                                    )
                                    + tab
                                )
                            ),
                            Assign(
                                targets=[
                                    Attribute(
                                        Name("argument_parser", Load()),
                                        "description",
                                        Store(),
                                    )
                                ],
                                value=set_value(
                                    (fill if wrap_description else identity)(
                                        intermediate_repr["doc"]
                                    )
                                ),
                                lineno=None,
                                expr=None,
                                **maybe_type_comment
                            ),
                        )
                    ),
                    filter(
                        None,
                        (
                            *(
                                (
                                    map(
                                        partial(
                                            param2argparse_param,
                                            word_wrap=word_wrap,
                                            emit_default_doc=emit_default_doc,
                                        ),
                                        intermediate_repr["params"].items(),
                                    )
                                )
                                if "params" in intermediate_repr
                                else ()
                            ),
                            *(
                                internal_body[
                                    2
                                    if len(internal_body) > 1
                                    and isinstance(internal_body[1], Assign)
                                    and internal_body[1].targets[0].id
                                    == "argument_parser"
                                    else 1 :
                                ]
                                if internal_body
                                and isinstance(internal_body[0], Expr)
                                and isinstance(get_value(internal_body[0].value), str)
                                else internal_body
                            ),
                            None
                            if internal_body and isinstance(internal_body[-1], Return)
                            else (
                                Return(
                                    value=Tuple(
                                        ctx=Load(),
                                        elts=[
                                            Name("argument_parser", Load()),
                                            set_value(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            if code_quoted(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            else ast.parse(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            .body[0]
                                            .value,
                                        ],
                                        expr=None,
                                    ),
                                    expr=None,
                                )
                                if "default"
                                in (
                                    intermediate_repr.get("returns")
                                    or {"return_type": iter(())}
                                )["return_type"]
                                else Return(
                                    value=Name("argument_parser", Load()), expr=None
                                )
                            ),
                        ),
                    ),
                )
            )
        ),
        decorator_list=[],
        name=function_name,
        returns=None,
        lineno=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )


def class_(
    intermediate_repr,
    emit_call=False,
    class_name="ConfigClass",
    class_bases=("object",),
    decorator_list=None,
    word_wrap=True,
    docstring_format="rest",
    emit_original_whitespace=False,
    emit_default_doc=False,
):
    """
    Construct a class

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param class_name: name of class
    :type class_name: ```str```

    :param class_bases: bases of class (the generated class will inherit these)
    :type class_bases: ```Iterable[str]```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param emit_original_whitespace: Whether to emit original whitespace or strip it out (in docstring)
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :returns: Class AST
    :rtype: ```ClassDef```
    """
    assert isinstance(intermediate_repr, dict), "{} != dict".format(
        type(intermediate_repr).__name__
    )
    returns = (
        intermediate_repr["returns"]
        if "return_type" in ((intermediate_repr or {}).get("returns") or iter(()))
        else OrderedDict()
    )

    param_names = frozenset(intermediate_repr["params"].keys())
    if returns:
        intermediate_repr["params"].update(returns)
        del intermediate_repr["returns"]

    internal_body = intermediate_repr.get("_internal", {}).get("body", [])
    # TODO: Add correct classmethod/staticmethod to decorate function using `annotate_ancestry` and first-field checks
    # Such that the `self.` or `cls.` rewrite only applies to non-staticmethods
    # assert internal_body, "Expected `internal_body` to have contents"
    if param_names:
        if internal_body:
            internal_body = list(
                map(
                    ast.fix_missing_locations,
                    map(RewriteName(param_names).visit, internal_body),
                )
            )
        elif (returns or {"return_type": None}).get("return_type") is not None:
            internal_body = returns["return_type"]

    indent_level = 1
    sep = indent_level * tab

    _emit_docstring = partial(
        docstring,
        docstring_format=docstring_format,
        indent_level=indent_level,
        emit_default_doc=emit_default_doc,
        emit_separating_tab=True,
        emit_types=False,
        word_wrap=word_wrap,
    )
    return ClassDef(
        bases=list(map(rpartial(Name, Load()), class_bases)),
        body=list(
            chain.from_iterable(
                (
                    (
                        Expr(
                            set_value(
                                "\n{sep}".format(sep=sep).join(
                                    (
                                        _emit_docstring(
                                            {
                                                "doc": intermediate_repr["doc"],
                                                "params": OrderedDict(),
                                                "returns": None,
                                            },
                                            emit_original_whitespace=emit_original_whitespace,
                                        ),
                                        _emit_docstring(
                                            {
                                                "doc": "",
                                                "params": intermediate_repr.get(
                                                    "params"
                                                ),
                                                "returns": intermediate_repr.get(
                                                    "returns"
                                                ),
                                            },
                                            emit_original_whitespace=False,
                                        )
                                        .replace(
                                            "\n{sep}:param ".format(sep=sep),
                                            ":cvar ",
                                        )
                                        .replace(
                                            "\n{sep}:returns:".format(sep=sep),
                                            ":cvar return_type:",
                                            1,
                                        )
                                        .rstrip(),
                                    )
                                ),
                            )
                        ),
                    ),
                    map(param2ast, intermediate_repr["params"].items()),
                    iter(
                        (
                            (
                                _make_call_meth(
                                    internal_body,
                                    returns["return_type"]["default"]
                                    if "default"
                                    in (
                                        (returns or {"return_type": iter(())}).get(
                                            "return_type"
                                        )
                                        or iter(())
                                    )
                                    else None,
                                    param_names,
                                    docstring_format=docstring_format,
                                    word_wrap=word_wrap,
                                ),
                            )
                            or iter(())
                        )
                        if emit_call and internal_body
                        else iter(())
                    ),
                )
            )
        ),
        decorator_list=list(map(rpartial(Name, Load()), decorator_list))
        if decorator_list
        else [],
        keywords=[],
        name=class_name,
        expr=None,
        identifier_name=None,
    )


def docstring(
    intermediate_repr,
    docstring_format="rest",
    purpose="function",
    word_wrap=True,
    indent_level=0,
    emit_separating_tab=True,
    emit_types=True,
    emit_original_whitespace=False,
    emit_default_doc=True,
):
    """
    Converts an IR to a docstring

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param purpose: Emit `:param` if purpose == 'function' elif purpose == 'class' then `:cvar`
    :type purpose: ```Literal['class', 'function']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param indent_level: indentation level whence: 0=no_tabs, 1=one tab; 2=two tabs
    :type indent_level: ```int```

    :param emit_separating_tab: Whether to put a tab between :param and return and desc
    :type emit_separating_tab: ```bool```

    :param emit_types: Whether to show `:type` lines
    :type emit_types: ```bool```

    :param emit_original_whitespace: Whether to emit original whitespace or strip it out
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :returns: docstring
    :rtype: ```str```
    """
    _sep = tab * indent_level
    return "\n{_sep}{}\n{_sep}".format(
        (
            emit_separating_tabs
            if emit_separating_tab and emit_original_whitespace is False
            else identity
        )(
            indent(
                "\n{doc}\n\n{nl0}{params}\n{returns}\n{nl1}".format(
                    doc=(fill if word_wrap else identity)(intermediate_repr["doc"]),
                    nl0="" if docstring_format == "rest" else "\n",
                    nl1="\n" if docstring_format == "numpydoc" else "",
                    params="\n{}".format(
                        "\n" if docstring_format == "rest" else ""
                    ).join(
                        (
                            lambda param_lines: [
                                getattr(ARG_TOKENS, docstring_format)[0]
                            ]
                            + param_lines
                            if param_lines and docstring_format != "rest"
                            else param_lines
                        )(
                            list(
                                map(
                                    partial(
                                        emit_param_str,
                                        style=docstring_format,
                                        purpose=purpose,
                                        emit_type=emit_types,
                                        emit_default_doc=emit_default_doc,
                                        word_wrap=word_wrap,
                                    ),
                                    intermediate_repr["params"].items(),
                                ),
                            )
                        )
                    ),
                    returns="".join(
                        (
                            lambda l: l
                            if l is None
                            else "{}\n{}".format(
                                ""
                                if docstring_format == "rest"
                                else "\n{}".format(
                                    getattr(RETURN_TOKENS, docstring_format)[0]
                                ),
                                l,
                            )
                        )(
                            next(
                                map(
                                    partial(
                                        emit_param_str,
                                        style=docstring_format,
                                        purpose=purpose,
                                        emit_type=emit_types,
                                        emit_default_doc=emit_default_doc,
                                        word_wrap=word_wrap,
                                    ),
                                    intermediate_repr["returns"].items(),
                                ),
                                None,
                            )
                        )
                    )
                    if "return_type" in (intermediate_repr.get("returns") or iter(()))
                    else "",
                ),
                _sep,
            ),
            indent_level=indent_level,
            run_per_line=identity,
        ).strip(),
        _sep=_sep,
    )


def file(node, filename, mode="a", skip_black=False):
    """
    Convert AST to a file

    :param node: AST node
    :type node: ```Union[Module, ClassDef, FunctionDef]```

    :param filename: emit to this file
    :type filename: ```str```

    :param mode: Mode to open the file in, defaults to append
    :type mode: ```str```

    :param skip_black: Whether to skip formatting with black
    :type skip_black: ```bool```

    :returns: None
    :rtype: ```NoneType```
    """
    if not isinstance(node, Module):
        node = Module(body=[node], type_ignores=[], stmt=None)
    src = to_code(node)
    if not skip_black:
        src = black.format_str(
            src,
            mode=black.Mode(
                target_versions=set(),
                line_length=119,
                is_pyi=False,
                string_normalization=False,
            ),
        )
    with open(filename, mode) as f:
        f.write(src)


def function(
    intermediate_repr,
    function_name,
    function_type,
    word_wrap=True,
    emit_default_doc=False,
    docstring_format="rest",
    indent_level=2,
    emit_separating_tab=PY3_8,
    type_annotations=True,
    emit_as_kwonlyargs=True,
    emit_original_whitespace=False,
):
    """
    Construct a function from our IR

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param function_name: name of function_def
    :type function_name: ```Optional[str]```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Optional[Literal['self', 'cls', 'static']]```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param indent_level: docstring indentation level whence: 0=no_tabs, 1=one tab; 2=two tabs
    :type indent_level: ```int```

    :param emit_separating_tab: docstring decider for whether to put a tab between :param and return and desc
    :type emit_separating_tab: ```bool```

    :param type_annotations: True to have type annotations (3.6+), False to place in docstring
    :type type_annotations: ```bool```

    :param emit_as_kwonlyargs: Whether argument(s) emitted must be keyword only
    :type emit_as_kwonlyargs: ```bool```

    :param emit_original_whitespace: Whether to emit an original whitespace (in docstring) or strip it out
    :type emit_original_whitespace: ```bool```

    :returns: AST node for function definition
    :rtype: ```FunctionDef```
    """
    params_no_kwargs = tuple(
        filter(
            lambda param: not param[0].endswith("kwargs"),
            intermediate_repr["params"].items(),
        )
    )

    function_name = function_name or intermediate_repr["name"]
    function_type = function_type or intermediate_repr["type"]

    args = (
        [] if function_type in frozenset((None, "static")) else [set_arg(function_type)]
    )
    from cdd.emitter_utils import ast_parse_fix

    args_from_params = list(
        map(
            lambda param: set_arg(
                annotation=(
                    Name(param[1]["typ"], Load())
                    if param[1]["typ"] in simple_types
                    else ast_parse_fix(param[1]["typ"])
                )
                if type_annotations and "typ" in param[1]
                else None,
                arg=param[0],
            ),
            params_no_kwargs,
        ),
    )
    defaults_from_params = list(
        map(
            lambda param: set_value(None)
            if param[1].get("default") in none_types
            else set_value(param[1].get("default")),
            params_no_kwargs,
        )
    )
    if emit_as_kwonlyargs:
        kwonlyargs, kw_defaults, defaults = args_from_params, defaults_from_params, []
    else:
        kwonlyargs, kw_defaults, defaults = [], [], defaults_from_params
        args += args_from_params

    internal_body = get_internal_body(
        target_name=function_name,
        target_type=function_type,
        intermediate_repr=intermediate_repr,
    )
    return_val = (
        Return(
            value=ast.parse(
                intermediate_repr["returns"]["return_type"]["default"].strip("`")
            )
            .body[0]
            .value,
            expr=None,
        )
        if (intermediate_repr.get("returns") or {"return_type": {}})["return_type"].get(
            "default"
        )
        else None
    )

    return FunctionDef(
        args=arguments(
            args=args,
            defaults=defaults,
            kw_defaults=kw_defaults,
            kwarg=next(
                map(
                    lambda param: set_arg(param[0]),
                    filter(
                        lambda param: param[0].endswith("kwargs"),
                        intermediate_repr["params"].items(),
                    ),
                ),
                None,
            ),
            kwonlyargs=kwonlyargs,
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=list(
            filter(
                None,
                (
                    Expr(
                        set_value(
                            docstring(
                                intermediate_repr,
                                docstring_format=docstring_format,
                                emit_default_doc=emit_default_doc,
                                emit_original_whitespace=emit_original_whitespace,
                                emit_separating_tab=emit_separating_tab,
                                emit_types=not type_annotations,
                                indent_level=indent_level,
                                word_wrap=word_wrap,
                            )
                        )
                    ),
                    *(
                        internal_body[:-1]
                        if internal_body
                        and isinstance(internal_body[-1], Return)
                        and return_val
                        else internal_body
                    ),
                    return_val,
                ),
            )
        ),
        decorator_list=[],
        name=function_name,
        returns=(
            ast.parse(intermediate_repr["returns"]["return_type"]["typ"]).body[0].value
            if type_annotations
            and (intermediate_repr.get("returns") or {"return_type": {}})[
                "return_type"
            ].get("typ")
            else None
        ),
        lineno=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )


def json_schema(
    intermediate_repr,
    identifier="https://offscale.io/json.schema.json",
    emit_original_whitespace=False,
):
    """
    Construct a JSON schema dict

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param identifier: The `$id` of the schema
    :type identifier: ```str```

    :param emit_original_whitespace: Whether to emit original whitespace (in top-level `description`) or strip it out
    :type emit_original_whitespace: ```bool```

    :returns: JSON Schema dict
    :rtype: ```dict```
    """
    required = []
    _param2json_schema_property = partial(param2json_schema_property, required=required)
    properties = dict(
        map(_param2json_schema_property, intermediate_repr["params"].items())
    )

    return {
        "$id": identifier,
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": deindent(
            add(
                *map(
                    partial(
                        docstring,
                        emit_default_doc=True,
                        emit_original_whitespace=emit_original_whitespace,
                        emit_types=True,
                    ),
                    (
                        {
                            "doc": intermediate_repr["doc"],
                            "params": OrderedDict(),
                            "returns": None,
                        },
                        {
                            "doc": "",
                            "params": OrderedDict(),
                            "returns": intermediate_repr["returns"],
                        },
                    ),
                )
            )
        ).lstrip("\n"),
        "type": "object",
        "properties": properties,
        "required": required,
    }


def sqlalchemy_table(
    intermediate_repr,
    name="config_tbl",
    docstring_format="rest",
    word_wrap=True,
    emit_original_whitespace=False,
    emit_default_doc=True,
):
    """
    Construct an `name = sqlalchemy.Table(name, metadata, Column(…), …)`

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param name: name of binding + table
    :type name: ```str```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param emit_original_whitespace: Whether to emit an original whitespace (in docstring) or strip it out
    :type emit_original_whitespace: ```bool```

    :param emit_original_whitespace: Whether to emit original whitespace or strip it out
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :returns: AST of the Table expression + assignment
    :rtype: ```ClassDef```
    """
    return Assign(
        targets=[Name(name, Store())],
        value=Call(
            func=Name("Table", Load()),
            args=list(
                chain.from_iterable(
                    (
                        iter(
                            (
                                set_value(name),
                                Name("metadata", Load()),
                            )
                        ),
                        map(
                            partial(param_to_sqlalchemy_column_call, include_name=True),
                            intermediate_repr["params"].items(),
                        ),
                    )
                )
            ),
            keywords=[
                keyword(
                    arg="comment",
                    value=set_value(
                        deindent(
                            add(
                                *map(
                                    partial(
                                        docstring,
                                        emit_default_doc=emit_default_doc,
                                        docstring_format=docstring_format,
                                        word_wrap=word_wrap,
                                        emit_original_whitespace=emit_original_whitespace,
                                        emit_types=True,
                                    ),
                                    (
                                        {
                                            "doc": intermediate_repr["doc"].lstrip()
                                            + "\n\n"
                                            if intermediate_repr["returns"]
                                            else "",
                                            "params": OrderedDict(),
                                            "returns": None,
                                        },
                                        {
                                            "doc": "",
                                            "params": OrderedDict(),
                                            "returns": intermediate_repr["returns"],
                                        },
                                    ),
                                )
                            ).strip()
                        )
                    ),
                    identifier=None,
                )
            ]
            if intermediate_repr["doc"]
            else [],
            expr=None,
            expr_func=None,
        ),
        lineno=None,
        expr=None,
        **maybe_type_comment
    )


def sqlalchemy(
    intermediate_repr,
    emit_repr=True,
    class_name="Config",
    class_bases=("Base",),
    decorator_list=None,
    table_name=None,
    docstring_format="rest",
    word_wrap=True,
    emit_original_whitespace=False,
    emit_default_doc=True,
):
    """
    Construct an SQLAlchemy declarative class

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_repr: Whether to generate a `__repr__` method
    :type emit_repr: ```bool```

    :param class_name: name of class
    :type class_name: ```str```

    :param class_bases: bases of class (the generated class will inherit these)
    :type class_bases: ```Iterable[str]```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param table_name: Table name, defaults to `class_name`
    :type table_name: ```str```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param emit_original_whitespace: Whether to emit an original whitespace (in docstring) or strip it out
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :returns: SQLalchemy declarative class AST
    :rtype: ```ClassDef```
    """
    return ClassDef(
        name=class_name,
        bases=list(map(lambda class_base: Name(class_base, Load()), class_bases)),
        decorator_list=decorator_list or [],
        keywords=[],
        body=list(
            filter(
                None,
                (
                    Expr(
                        set_value(
                            add(
                                *map(
                                    partial(
                                        docstring,
                                        docstring_format=docstring_format,
                                        emit_default_doc=emit_default_doc,
                                        emit_original_whitespace=emit_original_whitespace,
                                        emit_separating_tab=True,
                                        emit_types=True,
                                        indent_level=1,
                                        word_wrap=word_wrap,
                                    ),
                                    (
                                        {
                                            "doc": intermediate_repr["doc"],
                                            "params": OrderedDict(),
                                            "returns": None,
                                        },
                                        {
                                            "doc": "",
                                            "params": OrderedDict(),
                                            "returns": intermediate_repr["returns"],
                                        },
                                    ),
                                )
                            )
                        )
                    )
                    if intermediate_repr["doc"]
                    or intermediate_repr["returns"].get("return_type", {}).get("doc")
                    else None,
                    Assign(
                        targets=[Name("__tablename__", Store())],
                        value=set_value(table_name or class_name),
                        expr=None,
                        lineno=None,
                        **maybe_type_comment
                    ),
                    *map(
                        lambda param: Assign(
                            targets=[Name(param[0], Store())],
                            value=param_to_sqlalchemy_column_call(
                                param, include_name=False
                            ),
                            expr=None,
                            lineno=None,
                            **maybe_type_comment
                        ),
                        intermediate_repr["params"].items(),
                    ),
                    generate_repr_method(
                        intermediate_repr["params"], class_name, docstring_format
                    )
                    if emit_repr
                    else None,
                ),
            )
        ),
        expr=None,
        identifier_name=None,
    )


__all__ = [
    "argparse_function",
    "class_",
    "docstring",
    "file",
    "function",
    "sqlalchemy_table",
    "sqlalchemy",
]



"""
Source transformer module. Uses astor on Python < 3.9
"""

from ast import AsyncFunctionDef, ClassDef, FunctionDef, Module, get_docstring, parse
from importlib import import_module
from sys import version_info

from cdd.ast_utils import annotate_ancestry
from cdd.pure_utils import reindent, tab


def to_code(node):
    """
    Convert the AST input to Python source string

    :param node: AST node
    :type node: ```AST```

    :returns: Python source
    :rtype: ```str```
    """
    return (
        getattr(import_module("astor"), "to_source")
        if version_info[:2] < (3, 9)
        else getattr(import_module("ast"), "unparse")
    )(node)


def ast_parse(
    source,
    filename="<unknown>",
    mode="exec",
    skip_annotate=False,
    skip_docstring_remit=False,
):
    """
    Convert the AST input to Python source string

    :param source: Python source
    :type  source: ```str```

    :param filename: Filename being parsed
    :type filename: ```str```

    :param mode: 'exec' to compile a module, 'single' to compile a, single (interactive) statement,
      or 'eval' to compile an expression.
    :type mode: ```Literal['exec', 'single', 'eval']```

    :param skip_annotate: Don't run `annotate_ancestry`
    :type skip_annotate: ```bool```

    :param skip_docstring_remit: Don't parse & emit the docstring as a replacement for current docstring
    :type skip_docstring_remit: ```bool```

    :returns: AST node
    :rtype: node: ```AST```
    """
    parsed_ast = parse(source, filename=filename, mode=mode)
    if not skip_annotate:
        annotate_ancestry(parsed_ast)
    if not skip_docstring_remit and isinstance(
        parsed_ast, (Module, ClassDef, FunctionDef, AsyncFunctionDef)
    ):
        docstring = get_docstring(parsed_ast)
        if docstring is None:
            return parsed_ast

        # Reindent docstring
        parsed_ast.body[0].value.value = "\n{tab}{docstring}\n{tab}".format(
            tab=tab, docstring=reindent(docstring)
        )
    return parsed_ast


__all__ = ["ast_parse", "to_code"]
